/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __IPC_AGENT_H__
#define __IPC_AGENT_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.Sysprof.Agent */

#define IPC_TYPE_AGENT (ipc_agent_get_type ())
#define IPC_AGENT(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), IPC_TYPE_AGENT, IpcAgent))
#define IPC_IS_AGENT(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), IPC_TYPE_AGENT))
#define IPC_AGENT_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), IPC_TYPE_AGENT, IpcAgentIface))

struct _IpcAgent;
typedef struct _IpcAgent IpcAgent;
typedef struct _IpcAgentIface IpcAgentIface;

struct _IpcAgentIface
{
  GTypeInterface parent_iface;


  gboolean (*handle_force_exit) (
    IpcAgent *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_send_signal) (
    IpcAgent *object,
    GDBusMethodInvocation *invocation,
    gint arg_signum);

  void (*log) (
    IpcAgent *object,
    const gchar *arg_message);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (IpcAgent, g_object_unref)
#endif

GType ipc_agent_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *ipc_agent_interface_info (void);
guint ipc_agent_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void ipc_agent_complete_force_exit (
    IpcAgent *object,
    GDBusMethodInvocation *invocation);

void ipc_agent_complete_send_signal (
    IpcAgent *object,
    GDBusMethodInvocation *invocation);



/* D-Bus signal emissions functions: */
void ipc_agent_emit_log (
    IpcAgent *object,
    const gchar *arg_message);



/* D-Bus method calls: */
void ipc_agent_call_force_exit (
    IpcAgent *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean ipc_agent_call_force_exit_finish (
    IpcAgent *proxy,
    GAsyncResult *res,
    GError **error);

gboolean ipc_agent_call_force_exit_sync (
    IpcAgent *proxy,
    GCancellable *cancellable,
    GError **error);

void ipc_agent_call_send_signal (
    IpcAgent *proxy,
    gint arg_signum,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean ipc_agent_call_send_signal_finish (
    IpcAgent *proxy,
    GAsyncResult *res,
    GError **error);

gboolean ipc_agent_call_send_signal_sync (
    IpcAgent *proxy,
    gint arg_signum,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define IPC_TYPE_AGENT_PROXY (ipc_agent_proxy_get_type ())
#define IPC_AGENT_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), IPC_TYPE_AGENT_PROXY, IpcAgentProxy))
#define IPC_AGENT_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), IPC_TYPE_AGENT_PROXY, IpcAgentProxyClass))
#define IPC_AGENT_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), IPC_TYPE_AGENT_PROXY, IpcAgentProxyClass))
#define IPC_IS_AGENT_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), IPC_TYPE_AGENT_PROXY))
#define IPC_IS_AGENT_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), IPC_TYPE_AGENT_PROXY))

typedef struct _IpcAgentProxy IpcAgentProxy;
typedef struct _IpcAgentProxyClass IpcAgentProxyClass;
typedef struct _IpcAgentProxyPrivate IpcAgentProxyPrivate;

struct _IpcAgentProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  IpcAgentProxyPrivate *priv;
};

struct _IpcAgentProxyClass
{
  GDBusProxyClass parent_class;
};

GType ipc_agent_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (IpcAgentProxy, g_object_unref)
#endif

void ipc_agent_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
IpcAgent *ipc_agent_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
IpcAgent *ipc_agent_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void ipc_agent_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
IpcAgent *ipc_agent_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
IpcAgent *ipc_agent_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define IPC_TYPE_AGENT_SKELETON (ipc_agent_skeleton_get_type ())
#define IPC_AGENT_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), IPC_TYPE_AGENT_SKELETON, IpcAgentSkeleton))
#define IPC_AGENT_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), IPC_TYPE_AGENT_SKELETON, IpcAgentSkeletonClass))
#define IPC_AGENT_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), IPC_TYPE_AGENT_SKELETON, IpcAgentSkeletonClass))
#define IPC_IS_AGENT_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), IPC_TYPE_AGENT_SKELETON))
#define IPC_IS_AGENT_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), IPC_TYPE_AGENT_SKELETON))

typedef struct _IpcAgentSkeleton IpcAgentSkeleton;
typedef struct _IpcAgentSkeletonClass IpcAgentSkeletonClass;
typedef struct _IpcAgentSkeletonPrivate IpcAgentSkeletonPrivate;

struct _IpcAgentSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  IpcAgentSkeletonPrivate *priv;
};

struct _IpcAgentSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType ipc_agent_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (IpcAgentSkeleton, g_object_unref)
#endif

IpcAgent *ipc_agent_skeleton_new (void);


G_END_DECLS

#endif /* __IPC_AGENT_H__ */
